﻿using System.Collections;
using System.Collections.Generic;
using UnityEngine;

// 設定したエージェントをランダムに配置するなど．
// エージェントの管理．
public class AgentManager : MonoBehaviour
{
    // 時間幅
    [LockIf("lockOnPlay")]
    public float tau = 0.1f;
    [HideInInspector]
    public bool lockOnPlay = false;


    // エージェントの設定をまとめたクラス
    [System.Serializable]
    public class AgentSetting
    {
        // 複製するエージェントの原本
        public Transform original;

        // 複製するエージェントの数
        public int nAgent = 1;

        // ランダムに配置する位置の下限と上限
        public Vector3 initialPositionLowerBound = Vector3.zero;
        public Vector3 initialPositionUpperBound = Vector3.one;

        // 回転（向き）をランダムに配置するかどうか
        public bool randomizeRotation = true;

        // randomizeRotationがfalseの時のエージェントの向き
        public Vector3 initialDirection = Vector3.forward;
    }


    // 複数の種類のエージェントの設定
    [SerializeField]
    List<AgentSetting> agentSettings = new List<AgentSetting>(){new AgentSetting()};

    // インスタンス化したエージェントのリスト
    private List<List<Agent>> agents;


    void Start()
    {
        agents = new List<List<Agent>>();

        foreach(AgentSetting agentSetting in agentSettings)
        {
            if(agentSetting.original){
                AgentInstantiate(agentSetting);
            }
        }

        Time.fixedDeltaTime = tau;
        lockOnPlay = true;
    }

    
    // agentSettingをもとにエージェントを配置していく．
    void AgentInstantiate(AgentSetting agentSetting)
    {
        List<Agent> agents = new List<Agent>();

        if(agentSetting.randomizeRotation)
        {
            for(int i = 0; i < agentSetting.nAgent; i++)
            {
                Vector3 position = Vector3.zero;
                position.x = Random.Range(agentSetting.initialPositionLowerBound.x, agentSetting.initialPositionUpperBound.x);
                position.y = Random.Range(agentSetting.initialPositionLowerBound.y, agentSetting.initialPositionUpperBound.y);
                position.z = Random.Range(agentSetting.initialPositionLowerBound.z, agentSetting.initialPositionUpperBound.z);

                Quaternion rotation = Random.rotation;

                agents.Add(Instantiate(agentSetting.original, position, rotation).GetComponent<Agent>());
            }
        }
        else
        {
            for(int i = 0; i < agentSetting.nAgent; i++)
            {
                Vector3 position = Vector3.zero;
                position.x = Random.Range(agentSetting.initialPositionLowerBound.x, agentSetting.initialPositionUpperBound.x);
                position.y = Random.Range(agentSetting.initialPositionLowerBound.y, agentSetting.initialPositionUpperBound.y);
                position.z = Random.Range(agentSetting.initialPositionLowerBound.z, agentSetting.initialPositionUpperBound.z);
                
                Quaternion rotation = Quaternion.LookRotation(agentSetting.initialDirection);

                agents.Add(Instantiate(agentSetting.original, position, rotation).GetComponent<Agent>());
            }
        }

        // すべてのエージェントをactiveにする．
        // ついでに分かりやすい名前をつけておく．
        string agentTag = agentSetting.original.tag;
        string agentLayer = LayerMask.LayerToName(agentSetting.original.gameObject.layer);
        
        int agentCount = 0;
        foreach(Agent agent in agents)
        {
            agent.gameObject.SetActive(true);
            agent.transform.name = $"{agentTag} {agentLayer} ({agentCount})";
            agentCount++;
        }

        // エージェントのリストのリストに追加
        this.agents.Add(agents);
    }


    ////////////////////////////
    // 集団の特徴を計算する関数 //
    ////////////////////////////

    // 集団極性:集団の連帯度合い
    static public float GroupPolarization(List<Agent> agents)
    {
        int nAgents = 0;
        Vector3 v = Vector3.zero;

        foreach(Agent agent in agents)
        {
            v += agent.Velocity;
            nAgents++;
        }

        return v.magnitude / nAgents;
    }

    // 集団の重心
    static public Vector3 GroupCenter(List<Agent> agents)
    {
        int nAgents = 0;
        Vector3 c = Vector3.zero;

        foreach(Agent agent in agents)
        {
            c += agent.Position;
            nAgents++;
        }

        return c / nAgents;
    }

    // 集団角運動量：集団の重心まわりの回転度合い
    static public float GroupAngularMomentum(List<Agent> agents)
    {
        int nAgents = 0;
        Vector3 m = Vector3.zero;
        Vector3 c = GroupCenter(agents);

        foreach(Agent agent in agents)
        {
            m += Vector3.Cross(agent.Position - c, agent.Velocity);
            nAgents++;
        }

        return m.magnitude / nAgents;
    }
}
